# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

import os
import logging

from petaly.utils.file_handler import FileHandler
from petaly.core.f_loader import FLoader
from petaly.utils.utils import FormatDict
from petaly.connectors.gcp.gs.gs_connector import GSConnector


class GSLoader(FLoader):
    def __init__(self, pipeline):
        self.gs_connector = GSConnector()
        self.f_handler = FileHandler()

        super().__init__(pipeline)
        self.cloud_bucket_name = self.pipeline.target_attr.get('gcp_bucket_name')
        self.cloud_project_id = self.pipeline.target_attr.get('gcp_project_id')
        self.cloud_region = self.pipeline.target_attr.get('gcp_region')
        #self.cloud_service_account = self.pipeline.target_attr.get('gcp_service_account')

        self.cloud_bucket_path = self.gs_connector.bucket_prefix + self.cloud_bucket_name + '/'
        self.load_from_bucket = False if self.cloud_bucket_name is None else True

    def load_data(self):
        super().load_data()

    def load_from(self, loader_obj_conf):

        object_name = loader_obj_conf.get('object_name')

        self.gs_connector.delete_gs_folder(self.cloud_bucket_name, object_name)

        output_data_object_dir = loader_obj_conf.get('output_data_object_dir')
        self.f_handler.gzip_csv_files(output_data_object_dir, cleanup_file=True)
        file_list = self.f_handler.get_specific_files(output_data_object_dir, '*.csv.gz')

        if self.load_from_bucket == True:
            blob_prefix_dir = self.pipeline.target_attr.get('destination_blob_dir')
            bucket_file_list = self.load_files_to_gs(file_list, self.cloud_bucket_name, blob_prefix_dir, object_name)


    def deprecated_cleanup_cloud(self):
        """ function to recursively search for files named object_metadata.yaml in the pipeline's output directory.
        """

        dir_files = self.f_handler.get_specific_files(self.pipeline.output_pipeline_dpath, self.pipeline.object_metadata_fname)

        for file in dir_files:
            # Conduct Paths and Names
            self.gs_connector.delete_gs_folder(self.cloud_bucket_name, file)

        self.f_connector.delete_gs_folder(self.cloud_bucket_name, self.pipeline.pipeline_name)

    def deprecated_load_file_to_gs(self, file_local_fpath, cloud_bucket_name, object_name ):
        """ upload file to GS bucket
        """

        file_name = os.path.basename(file_local_fpath)

        blob_prefix_dir = cloud_bucket_name
        if blob_prefix_dir is None:
            blob_prefix_dir = ''

        blob_dir = blob_prefix_dir + '/'+ self.pipeline.pipeline_name + "/" + object_name

        blob_path = blob_dir + "/" + file_name
        self.gs_connector.upload_blob(file_local_fpath, cloud_bucket_name, blob_path)
        return self.cloud_bucket_path + blob_path

    def load_files_to_gs(self, local_file_list, cloud_bucket_name, blob_prefix_dir, object_name):
        """ upload file to GS bucket
        """
        bucket_file_list = []
        for file_local_fpath in local_file_list:
            file_name = os.path.basename(file_local_fpath)

            if blob_prefix_dir is None:
                blob_prefix_dir = ''

            blob_prefix_dir = blob_prefix_dir.strip('/')

            blob_path = blob_prefix_dir + '/' + object_name + '/' + file_name
            self.gs_connector.upload_blob(file_local_fpath, cloud_bucket_name, blob_path)

            full_blob_path = self.gs_connector.bucket_prefix + cloud_bucket_name + '/' + blob_path
            bucket_file_list.append(self.gs_connector.bucket_prefix + cloud_bucket_name + '/' + blob_path)

            logging.debug(f"File upload: {full_blob_path}")

        return bucket_file_list


# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import logging
logger = logging.getLogger(__name__)

class DataObject:
    def __init__(self, pipeline, object_name):

        data_objects = pipeline.data_objects_spec
        self.pipeline_data_object_dir = pipeline.output_object_data_dpath.format(object_name=object_name)
        self.data_objects_spec_mode = pipeline.data_attributes.get('data_objects_spec_mode')
        self.object_settings = self.format_object_default_settings(pipeline.object_default_settings)
        self.object_settings.update({'cleanup_linebreak_in_fields': False})

        data_object_spec = self.get_object_spec(data_objects, object_name)
        logger.debug(f"Data object spec: {data_object_spec}")
        if data_object_spec == {}:

            if self.data_objects_spec_mode == 'only':
                logger.info(
                    f"For {pipeline.source_connector_id} extract the parameters data_objects_spec_mode=only and specification in the data_objects_spec[] are required. Use python -m petaly init -p {pipeline.pipeline_name} --object_name table1,table2 -c your_config_dir/petaly.ini")
                sys.exit()

            elif self.data_objects_spec_mode in ('ignore', 'prefer'):
                if pipeline.source_connector_id in ('csv'):
                    logger.info(
                        f"In case your source is csv, the parameters data_objects_spec_mode should be set to only and require the specification in the data_objects_spec[]."
                        f"\ndata_objects_spec_mode=only"
                        f"\nCheck pipeline under: {pipeline.pipeline_fpath}")

                    sys.exit()

            return self.set_default_object_spec(pipeline, object_name)

        self.object_name = data_object_spec.get('object_spec').get('object_name')
        self.destination_object_name = data_object_spec.get('object_spec').get('destination_object_name')
        self.recreate_destination_object = True if data_object_spec.get('object_spec').get('recreate_destination_object') is True else False
        self.exclude_columns = data_object_spec.get('object_spec').get('exclude_columns')
        self.object_source_dir = data_object_spec.get('object_spec').get('object_source_dir')
        self.file_names = data_object_spec.get('object_spec').get('file_names')
        cleanup_linebreak_in_fields = data_object_spec.get('object_spec').get('cleanup_linebreak_in_fields')
        self.object_settings.update({'cleanup_linebreak_in_fields': cleanup_linebreak_in_fields})

    def to_dict(self) -> dict:
        return {key: value for key, value in self.__dict__.items()}

    def get_object_spec(self, data_objects, object_name):

        return_object_spec = {}
        for object_spec in data_objects.get('data_objects_spec'):
            if object_spec != None:
                if object_spec.get('object_spec').get('object_name') == object_name:
                    return_object_spec = object_spec.copy()
        return return_object_spec

    def set_default_object_spec(self, pipeline, object_name):
        self.object_name = object_name
        self.destination_object_name = None
        self.recreate_destination_object = False
        self.cleanup_linebreak_in_fields = False
        self.exclude_columns = [None]
        self.object_source_dir = None
        self.file_names = [None]

    def format_object_default_settings(self, object_default_settings):
        """
        """
        object_settings = object_default_settings.copy()
        # 1. Header

        header = True if object_settings.get('header') is True else False
        object_settings.update({'header': header})

        # 2. Columns delimiter
        columns_delimiter = object_settings.get('columns_delimiter')
        #object_settings.update({'columns_delimiter_origin': columns_delimiter})
        if columns_delimiter == "\\t":
            object_settings.update({'columns_delimiter': '\t'})

        return object_settings



# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

from petaly.utils.file_handler import FileHandler
from petaly.core.pipeline import Pipeline
from petaly.cli.cli_menu import CliMenu
from petaly.core.composer import Composer
import rich.prompt as prompt


class CliInitializer():

	def __init__(self, main_config):
		self.m_conf = main_config
		self.cli_menu = CliMenu(main_config)
		self.console = self.cli_menu.console
		self.composer = Composer()
		self.f_handler = FileHandler()

	### Init Workspace
	def init_workspace(self, skip_message_if_exist=False):
		"""
		"""
		self.create_dir(self.m_conf.pipeline_base_dpath, "Pipeline base", skip_message_if_exist)
		self.create_dir(self.m_conf.output_base_dpath, "Output base", skip_message_if_exist)
		self.create_dir(self.m_conf.logs_base_dpath, "Logs base", skip_message_if_exist)

	def create_dir(self, dpath, dir_type_message, skip_message_if_exist):

		if self.f_handler.is_dir(dpath):
			if not skip_message_if_exist:
				self.console.print (f"{dir_type_message} directory {dpath} already exists.")
		else:
			self.f_handler.make_dirs(dpath)
			self.console.print(f"{dir_type_message} directory {dpath} is created.")


	### Init Pipeline
	def init_pipeline(self, pipeline_name):
		""" Initiate sub folder under pipeline directory with given pipeline_name
		"""
		self.console.print(f"[bold]{self.cli_menu.break_line}[/bold]")
		self.console.print(f"[bold]The pipeline initialization has started.[/bold]\n")

		if pipeline_name is None:
			pipeline_name = self.cli_menu.force_assign_value(key='pipeline_name', message="Specify unique pipeline name")

		self.cli_menu.composed_pipeline_config[0]['pipeline']['pipeline_attributes'].update({'pipeline_name': pipeline_name})

		pipeline_dpath = os.path.join(self.m_conf.pipeline_base_dpath, pipeline_name)
		pipeline_fpath = os.path.join(pipeline_dpath, self.m_conf.pipeline_fname)
		output_pipeline_dpath = os.path.join(self.m_conf.output_base_dpath, pipeline_name)

		# check pipeline directory and pipeline.yaml
		if self.f_handler.is_file(pipeline_fpath):
			self.console.print (f"Pipeline with the path {pipeline_dpath} already exists.")
			process_continue = self.cli_menu.prompt.Confirm.ask(f"\nDo you want to continue and overwrite the existing {self.m_conf.pipeline_fname} configuration?")

			if process_continue:
				self.console.print(f"Backup with the name {self.m_conf.pipeline_fname}.buckup_<timestamp> from pipeline.yaml will be created.")
			else:
				sys.exit()

		elif self.f_handler.is_dir(pipeline_dpath):
			pass
		else:
			self.f_handler.make_dirs(pipeline_dpath)

		# check output folder and ask for confirmation
		if self.f_handler.is_dir(output_pipeline_dpath):
			self.console.print (f"\nOutput directory with the path {output_pipeline_dpath} already exists.")
			process_continue = self.cli_menu.prompt.Confirm.ask(f"\nDo you want to continue and overwrite the existing output directory? All files inside will be deleted.")
			if process_continue is True:
				self.f_handler.cleanup_dir(output_pipeline_dpath)
				self.f_handler.make_dirs(output_pipeline_dpath)
			else:
				sys.exit()

		else:
			self.f_handler.make_dirs(output_pipeline_dpath)

		self.cli_menu.compose_pipeline(pipeline_name)
		self.f_handler.backup_file(pipeline_fpath)
		self.f_handler.save_dict_to_yaml(pipeline_fpath, self.cli_menu.composed_pipeline_config, dump_all=True)

		self.console.print(f"\nCheck pipeline {pipeline_name} under: {pipeline_fpath}")
		self.console.print(f"Check output directory under: {output_pipeline_dpath}")

		data_objects_spec_mode = self.cli_menu.composed_pipeline_config[0]['pipeline']['data_attributes'].get('data_objects_spec_mode')

		process_continue = True
		if data_objects_spec_mode in ("ignore"):
			self.console.print(f"\nThe parameter [bold]data_objects_spec_mode[/bold] was set to [bold]ignore[/bold]")
			process_continue = self.cli_menu.prompt.Confirm.ask(f"Do you want to continue defining the data objects despite the definition in data_objects_spec_mode=ignore?")

		if process_continue:
			self.console.print("\nContinue with the configuration of objects/tables in the next step.")

			self.init_data_objects(pipeline_name, object_names=None)

		else:
			self.console.print("Review the pipeline.yaml file and modify manually if necessary.")

	def init_data_objects(self, pipeline_name, object_names):

		if pipeline_name is None:
			pipeline_name = self.cli_menu.force_assign_value(key='pipeline_name',
													message="Provide pipeline name. Pipeline with this name should already exists.")

		pipeline = Pipeline(pipeline_name, self.m_conf)
		pipeline_all_obj = pipeline.get_pipeline_entire_config()
		data_objects_spec_meta = self.cli_menu.pipeline_meta_config.get('data_objects_spec')
		object_name_key_comment = data_objects_spec_meta.get('object_name').get("key_comment")
		data_objects_spec_list = []

		use_pipeline_wizard = prompt.Confirm.ask("Use data object specifications wizard")
		ask_for_next_object = True
		# 1. first handle list in array
		if object_names is not None and type(object_names) == str:

			object_name_arr = [item.strip() for item in object_names.split(',')]

			for i, object_name in enumerate(object_name_arr):
				self.console.print(f"\n[bold]{self.cli_menu.break_line}[/bold]")
				self.console.print(f"Continue with the object [bold]{object_name}[/bold] specification")

				object_spec = self.cli_menu.compose_object_spec(pipeline, object_name=object_name, use_pipeline_wizard=use_pipeline_wizard)

				data_objects_spec_list.append(object_spec)
				self.composer.save_data_objects(pipeline_all_obj=pipeline_all_obj, data_objects_spec=data_objects_spec_list, pipeline_fpath=pipeline.pipeline_fpath)

			ask_for_next_object = False

		# 2. continues here if array is emtpy or was handel and adding_objects is still true
		while ask_for_next_object:
			object_name = self.cli_menu.force_assign_value(key='object_name', message=object_name_key_comment)

			object_spec = self.cli_menu.compose_object_spec(pipeline, object_name=object_name, use_pipeline_wizard=use_pipeline_wizard)
			data_objects_spec_list.append(object_spec)

			# save each object to the yaml document
			self.composer.save_data_objects(pipeline_all_obj=pipeline_all_obj, data_objects_spec=data_objects_spec_list, pipeline_fpath=pipeline.pipeline_fpath)

			self.console.print(f"\n[bold]{self.cli_menu.break_line}[/bold]")
			self.console.print(f"The object: [bold]{object_name}[/bold] has been added to the pipeline.")

			ask_for_next_object = prompt.Confirm.ask("\nDo you want to continue defining the next data object?")

		if len(data_objects_spec_list) > 0:
			self.console.print(f"Data-Objects were added for pipeline {pipeline.pipeline_name}. For further configuration review the yaml file: {pipeline.pipeline_fpath} ")
		else:
			self.console.print(
				f"Data-Objects weren't specified for pipeline {pipeline.pipeline_name}. For further configuration review the yaml file: {pipeline.pipeline_fpath} ")



# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
logger = logging.getLogger(__name__)

from pprint import pprint

import time
from functools import wraps

def measure_time(func):
    """This decorator return the execution time for the decorated function."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        logger.debug(f"Function {func.__module__}.{func.__name__}()  ran in {round(end - start, 2)}s")

        return result

    return wrapper

class FormatDict(dict):
    """ With help of this class the function str.format_map() will ignore a key which wasn't specified in the parameter section

    """
    def __missing__(self, key):
        return '{' + str(key) + '}'

def ptprint(msg):
    pprint(msg)


# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
logger = logging.getLogger(__name__)

import os
import sys
import time
from petaly.core.f_loader import FLoader


class CsvLoader(FLoader):

    def __init__(self, pipeline):
        self.file_format = 'csv'
        super().__init__(pipeline)

    def load_data(self):
        super().load_data()

    def load_from(self, loader_obj_conf):

        object_name = loader_obj_conf.get('object_name')

        output_object_dir = self.pipeline.output_object_data_dpath.format(object_name=object_name)
        data_object = super().get_data_object(object_name)

        dest_file_dir = self.pipeline.target_attr.get("destination_file_dir")
        if dest_file_dir is None:
            logger.warning(f"The pipeline->target_attribute->destination_file_dir in pipeline.yaml is not specified.")
            sys.exit()

        dest_object_name = object_name
        if data_object.destination_object_name is not None:
            dest_object_name = data_object.destination_object_name

        # get target file directory destination_file_dir/pipeline_name/object_name
        dest_file_dpath = os.path.join(dest_file_dir, self.pipeline.pipeline_name, dest_object_name)

        # get target file format
        logger.debug(f"Destination file format: {self.file_format}")

        dir_exists, files_in_dir = self.f_handler.check_dir(output_object_dir)

        if dir_exists is True and files_in_dir> 0:

            logger.debug(f"Output file dir: {output_object_dir}")
            logger.info(f"Load object: {object_name} destination directory: {dest_file_dpath}")

            # lists file with .csv extension, also files ends with .csv.gz. The logic can be improved.
            file_list = self.f_handler.get_file_names_with_extensions(output_object_dir, self.file_format,  '.gz')

            for file in file_list:
                logger.debug(f"File: {file}")
                file_source_fpath = os.path.join(output_object_dir, file)

                # Rename file if destination_object_name is different from object_name
                dest_file_name = file.replace(object_name, dest_object_name)

                logger.debug(f"Load: Upload performed to destination path: {os.path.join(dest_file_dpath, dest_file_name)}")
                self.f_handler.cp_file(file_source_fpath, dest_file_dpath, target_file_name=dest_file_name)

        else:
            logger.error(f"Load object {object_name} failed. Check the source and pipeline.yaml configuration. "
                         f"Output directory doesn't exist or is empty: {output_object_dir}")




# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

from abc import ABC, abstractmethod

import os
import sys
import time
from pyarrow import csv, parquet, lib as pyarrow_lib

from petaly.utils.file_handler import FileHandler
from petaly.core.object_metadata import ObjectMetadata
from petaly.core.data_object import DataObject


class FExtractor(ABC):

    def __init__(self, pipeline):
        self.pipeline = pipeline
        self.f_handler = FileHandler()
        self.object_metadata = ObjectMetadata(pipeline)
        self.object_default_settings = pipeline.data_attributes.get("object_default_settings")
        pass

    @abstractmethod
    def extract_to(self, extractor_obj_conf):
        pass

    def extract_data(self):
        """ Its export data as csv into pipeline output directory.
        """

        logger.info(f"[--- Extract from {self.pipeline.source_connector_id} ---]")
        start_total_time = time.time()

        # 1. Start with cleanup
        #self.f_handler.cleanup_dir(self.pipeline.output_pipeline_dpath)

        # 4. save metadata and export scripts
        object_list = self.pipeline.data_objects

        # 5. run loop for each object
        for object_name in object_list:
            logger.info(f"Extract object: {object_name} started...")
            start_time = time.time()

            extractor_obj_conf = self.get_extractor_obj_conf(object_name)
            self.extract_to(extractor_obj_conf)
            end_time = time.time()
            logger.info(f"Extract object: {object_name} completed | time: {round(end_time - start_time, 2)}s")

        end_total_time = time.time()
        logger.info(f"Extract completed, duration: {round(end_total_time - start_total_time, 2)}s")

    def get_extractor_obj_conf(self, object_name) -> dict:

        extractor_obj_conf = {'object_name': object_name}

        # 1. compose output data dir
        output_metadata_object_dir = self.pipeline.output_object_metadata_dpath.format(object_name=object_name)
        extractor_obj_conf.update({'output_metadata_object_dir': output_metadata_object_dir})

        # 2. get and compose metadata query
        #metadata_fpath = self.pipeline.output_object_metadata_fpath.format(object_name=object_name)
        #table_metadata = self.f_handler.load_file_as_dict(metadata_fpath, 'json')
        #extract_queries_dict = self.compose_extract_queries(table_metadata)
        #extractor_obj_conf.update(extract_queries_dict)

        # 3. add object_default_settings
        # extractor_obj_conf.update({'object_settings': table_metadata.get('object_settings')})
        data_object = self.get_data_object(object_name)

        if data_object.object_source_dir is None:
            logger.error(f"Incorrect object specification in file: {self.pipeline.pipeline_fpath} "
                           f"\ndata_objects_spec: "
                           f"\n- object_spec:"
                           f"\n    object_name: {object_name}"
                           f"\n    object_source_dir: IS EMPTY")
            sys.exit()

        extractor_obj_conf.update({'object_source_dir': data_object.object_source_dir})
        extractor_obj_conf.update({'file_names': data_object.file_names})

        logger.debug(f"The object settings combined with default settings: {data_object.object_settings}")
        extractor_obj_conf.update({'object_settings': data_object.object_settings})

        # 6. create output object dir and output_file_path
        output_object_dpath = self.pipeline.output_object_data_dpath.format(object_name=object_name)
        extractor_obj_conf.update({'output_object_dpath': output_object_dpath})

        self.f_handler.make_dirs(output_object_dpath)

        output_object_fpath = os.path.join(output_object_dpath, object_name + '.csv')
        extractor_obj_conf.update({'output_object_fpath': output_object_fpath})

        logger.debug(f"Config for data extract: {extractor_obj_conf}")
        return extractor_obj_conf

    def deprecated_get_local_output_path(self, object_name):

        object_dpath = self.pipeline.output_object_data_dpath.format(object_name=object_name)
        self.f_handler.make_dirs(object_dpath)
        output_object_fpath = os.path.join(object_dpath, object_name + '.csv')
        return output_object_fpath

    def save_metadata_into_file(self,meta_table):
        """ Its save a table result as a metadata into file
        """
        object_name = meta_table.get('source_object_name')
        source_object_fpath = self.pipeline.output_object_metadata_fpath.format(object_name=object_name)
        logger.debug(f"Format and save metadata for table {meta_table.get('source_object_name')} in {source_object_fpath}")
        self.f_handler.save_dict_to_file(source_object_fpath, meta_table, 'json')

    def extract_metadata_from_parquet_file(self, parquet_fpath):
        """ Its extract metadata from parquet file and also change the dict from parquet.read_metadata in more readable format.
        """

        pq_metadata = parquet.read_metadata(parquet_fpath).to_dict()
        pq_meta_columns = pq_metadata.get('row_groups')[0].get('columns')
        column_names = parquet.read_schema(parquet_fpath).names
        column_data_types = parquet.read_schema(parquet_fpath).types

        pq_columns_metadata_arr = []

        for i in range(len(pq_meta_columns)):
            column_dict = {}
            c_idx = column_names.index(pq_meta_columns[i].get('path_in_schema'))

            # column name
            column_name = pq_meta_columns[i].get('path_in_schema')
            column_dict.update({'column_name': column_name})

            # column type
            column_type = str(column_data_types[c_idx])
            if column_type == 'null':
                column_type = 'string'
            column_dict.update({'data_type': str(column_data_types[c_idx])})

            column_dict.update({'physical_type': str(pq_meta_columns[i].get('physical_type'))})
            column_dict.update({'is_stats_set': str(pq_meta_columns[i].get('is_stats_set'))})
            column_dict.update({'encodings': str(pq_meta_columns[i].get('encodings'))})
            column_dict.update({'compression': str(pq_meta_columns[i].get('compression'))})

            if pq_meta_columns[i].get('is_stats_set') == True:

                column_dict.update({'statistics_distinct_count': str(pq_meta_columns[i].get('statistics').get('distinct_count')) })
                column_dict.update({'statistics_min': str(pq_meta_columns[i].get('statistics').get('min'))})
                column_dict.update({'statistics_max': str(pq_meta_columns[i].get('statistics').get('max'))})
                column_dict.update({'statistics_distinct_count': str(pq_meta_columns[i].get('statistics').get('distinct_count'))})
                column_dict.update({'statistics_null_count': str(pq_meta_columns[i].get('statistics').get('null_count'))})

            else:
                column_dict.update({'statistics_distinct_count': None})
                column_dict.update({'statistics_min': None})
                column_dict.update({'statistics_max': None})
                column_dict.update({'statistics_num_values': None})
                column_dict.update({'statistics_null_count': None})


            pq_columns_metadata_arr.append(column_dict)

        return pq_columns_metadata_arr

    def compose_metadata_file(self, parquet_fpath, object_name):
        """ Its creates a metadata file with all the attributes needed to recreate a table on the target.
        """
        pq_columns_metadata_arr = self.extract_metadata_from_parquet_file(parquet_fpath)

        if len(pq_columns_metadata_arr) == 0:
            logger.warning("The process has failed to extract the metadata from the file, the column definition is empty.")
            sys.exit()
        meta_table = self.object_metadata.compose_object_meta_from_file(object_name, pq_columns_metadata_arr)
        return meta_table

    def describe_parquet_metadata(self, parquet_fpath):
        """ Its describe a parquet metadata from parquet file to stdout
        """
        # extract metadata from parquet file
        pq_columns_metadata_arr = self.extract_metadata_from_parquet_file(parquet_fpath)
        header_str = ""
        col_separator = " | "
        output_text = ""

        # exit if array is empty
        if len(pq_columns_metadata_arr) == 0:
            logger.warning("pq_columns_metadata_arr is empty.")
            sys.exit()

        # create the header line
        for key in pq_columns_metadata_arr[0]:
            header_str += key + col_separator
        output_text += header_str + '\n'
        output_text += '-'*len(header_str)  + '\n'

        # compose output row by row
        for i in range(len(pq_columns_metadata_arr)):
            row_str = ""
            for key in pq_columns_metadata_arr[i]:
                row_str += str(pq_columns_metadata_arr[i].get(key)) + col_separator
            output_text += row_str + '\n'
        return output_text

    def analyse_file_structure(self, output_source_file, object_name, file_format_extension):
        """ Analyse the csv file to determine the column format.
        Transform the source file to parquet format to read the column data type.
        Finally, store the result of this discovery to the metadata file
        """

        is_file = self.f_handler.is_file(output_source_file)
        if is_file == False:
            logger.error(f"Output source file {output_source_file} doesn't exists")
            sys.exit()

        if  self.f_handler.check_file_extension(output_source_file, file_format_extension) == False:
            logger.error(f"File format doesn't match")
            sys.exit()

        # transform csv to parquet
        parse_options = csv.ParseOptions(delimiter=self.object_default_settings.get("columns_delimiter"))

        try:
            logger.debug(f"Start reading the csv file: {output_source_file}")
            file_data = csv.read_csv(output_source_file, parse_options=parse_options)
        except pyarrow_lib.ArrowInvalid as err:
            logger.error(f"Error {err}")
            logger.info(f"Check that the {output_source_file} file matches the parsing options: {self.object_default_settings}")
            sys.exit()

        pq_file_name = self.f_handler.replace_file_extension(output_source_file, '.parquet')
        parquet_fpath = os.path.join(self.pipeline.output_object_data_dpath.format(object_name=object_name), pq_file_name)
        parquet.write_table(file_data, parquet_fpath)

        return parquet_fpath

    def get_data_object(self, object_name):
        return DataObject(self.pipeline, object_name)


# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys

from petaly.utils.file_handler import FileHandler


class Composer():

	def __init__(self):
		self.f_handler = FileHandler(file_format='json')
		#self.pipeline = pipeline
		pass


	def normalise_column_name(self, column_name):
		""" It replaces unusable characters in column name. e.g. ':' will be replaced with '_'
		"""
		column_name = column_name.replace(':', '_')
		column_name = column_name.replace('.', '_')

		return column_name

	def get_object_list_from_output_dir(self, pipeline):
		""" This function check if object defined in the second list pipeline_data_objects in pipeline.yaml -> data_objects_spec corresponds
				with the first list data_objects which is usually a folder structure in output directory.
		"""
		object_dir_list = self.f_handler.get_all_dir_names(pipeline.output_pipeline_dpath)
		pipeline_object_list = pipeline.data_objects

		if pipeline.data_objects_spec_mode in ("ignore","prefer"):
			return_list = object_dir_list
		else:
			return_list = self.get_data_objects_intersection(object_dir_list, pipeline_object_list)

		return return_list
	def get_data_objects_intersection(self, first_list, second_list):
		"""
		"""
		return_list = []

		for value in first_list:
			if value in second_list:
				return_list.append(value)

		return return_list

	def save_data_objects(self, pipeline_all_obj, data_objects_spec, pipeline_fpath):

		data_object_list = []
		# make a list of new added objects with the same index order
		for idx, obj in enumerate(data_objects_spec):
			data_object_list.insert(idx, obj.get('object_spec').get('object_name'))

		# if object_name is in new in the list replace it in the pipeline, else do nothing
		if pipeline_all_obj[1].get('data_objects_spec') is not None:
			if len(pipeline_all_obj[1].get('data_objects_spec')) > 0:
				if pipeline_all_obj[1].get('data_objects_spec')[0] is not None:

					for idx, object_spec in enumerate(pipeline_all_obj[1].get('data_objects_spec')):
						if object_spec.get('object_spec').get('object_name') in data_object_list:
							ind, obj = self.get_object_spec_from_array(data_objects_spec,
																	   object_spec.get('object_spec').get(
																		   'object_name'))
							pipeline_all_obj[1].get('data_objects_spec')[idx] = obj
							data_objects_spec.pop(ind)
							data_object_list.pop(ind)

		# add the entire new objects at the end of the pipeline
		if len(data_object_list) > 0:
			for i, obj in enumerate(data_objects_spec):
				pipeline_all_obj[1].get('data_objects_spec').append(obj)

		self.f_handler.backup_file(pipeline_fpath)
		self.f_handler.save_dict_to_yaml(pipeline_fpath, pipeline_all_obj, dump_all=True)

	def get_object_spec_from_array(self, data_objects_spec, object_name):

		for idx, obj_spec in enumerate(data_objects_spec):

			if object_name == obj_spec.get('object_spec').get('object_name'):
				return idx, obj_spec





# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

import os
import re
import pathlib
import sys

import yaml
from yaml import SafeDumper, Dumper
import json
import shutil
import ast
import gzip
from datetime import datetime


class IndentDumper(yaml.Dumper):
    """ This is a helper class to indent yaml file and make it more readable
    This class is used in FileHandler class. e.g. yaml.dump_all( dict_data, file, Dumper=IndentDumper,  sort_keys=False, allow_unicode=True)
    """
    def increase_indent(self, flow=False, indentless=False):
        return super(IndentDumper, self).increase_indent(flow, False)


class FileHandler:
    """ FileHandler is a file manager, to save, load, modify yaml, json, text files
    """
    def __init__(self, file_format="yaml"):
        """
        """
        if file_format != "yaml" and file_format != "json":  # error!
            raise ValueError("The file format {file_format} is not supported. Parameter expect yaml or json as a file format.".format(file_format=file_format))
        self.file_format = "json" if file_format == "json" else "yaml"
        self.file_extension = "." + self.file_format

    def load_file_as_dict(self, file_fpath, file_format):
        dict_data = {}
        if file_format == 'yaml':
            dict_data = self.load_yaml(file_fpath)
        elif file_format == 'json':
            dict_data = self.load_json(file_fpath)
        return dict_data

    def load_json(self, file_fpath) -> {}:
        try:
            with (open(file_fpath, mode="rt", encoding="utf-8") as f):
                obj = json.loads(f.read())
                return obj
        except Exception as e:
            logger.error(e)
            return {}

    def load_yaml_all(self, file_fpath):
        try:
            with (open(file_fpath, mode="rt", encoding="utf-8") as f):
                yaml_data = yaml.safe_load_all(f)
                obj = list(yaml_data)
                return obj
        except Exception as e:
            logger.error(e)
            sys.exit()

    def load_yaml(self, file_fpath):

        if self.is_file(file_fpath) != True:
            return None

        try:
            with (open(file_fpath, mode="rt", encoding="utf-8") as f):
                obj = yaml.safe_load(f)
                return obj
        except Exception as e:
            logger.error(e)
            sys.exit()
            #return {}


    def replace_file_extension(self, file_fpath, new_file_extension):
        """ This function replace extension from file_fpath to the new_file_extension
        """
        return pathlib.Path(file_fpath).with_suffix(new_file_extension)

    def get_file_extensions(self, file_fpath):
        """ This function replace extension from file_fpath to the new_file_extension
        """
        return pathlib.Path(file_fpath).suffixes

    def check_file_extension(self, file_fpath, file_extension):
        """
        """
        path_in_arr = os.path.splitext(file_fpath)
        return path_in_arr[1] == file_extension

    def save_dict_to_file(self, file_fpath, dict_data, file_format='yaml'):
        """
        """
        if file_format == 'yaml':
            self.save_dict_to_yaml(file_fpath, dict_data)
        elif file_format == 'json':
            self.save_dict_to_json(file_fpath, dict_data)

    def save_dict_to_yaml(self, file_fpath, dict_data, dump_all=False):
        """
        """
        result = 0
        file_fpath = self.replace_file_extension(file_fpath, ".yaml")
        dir_name = os.path.dirname(file_fpath)
        self.make_dirs(dir_name)

        # This block, in combination with yaml.safe_dump, replaces null values with empty values when saving to a YAML file.
        SafeDumper.add_representer(
            type(None),
            lambda dumper, value: dumper.represent_scalar(u'tag:yaml.org,2002:null', '')
        )

        logger.debug(f"Save dict as yaml under: {file_fpath}")

        if dump_all==False:

            with (open(file_fpath, mode="w", encoding="utf-8") as file):
                ## yaml.dump(dict_data, file, Dumper=IndentDumper, sort_keys=False, allow_unicode=True, default_flow_style=False)
                yaml.safe_dump(dict_data, file, sort_keys=False, allow_unicode=True, default_flow_style=False, default_style=None)
        else:
            with (open(file_fpath, mode="w", encoding="utf-8") as file):
                ## yaml.dump_all(dict_data, file, sort_keys=False, allow_unicode=True, default_flow_style=False)
                yaml.safe_dump_all(dict_data, file, sort_keys=False, allow_unicode=True, default_flow_style=False, default_style=None)


    def save_dict_to_json(self, file_fpath, dict_data):
        """
        """
        result = 0
        logger.debug(f"Save dict as json under: {file_fpath}")

        file_fpath = self.replace_file_extension(file_fpath, ".json")
        dir_name = os.path.dirname(file_fpath)
        self.make_dirs(dir_name)

        with (open(file_fpath, mode="w", encoding="utf-8") as file):
            json_object = json.dumps(dict_data, indent=2, separators=(',', ':'))
            file.write(json_object)


    def format_json_file_compact(self, file_fpath, modify_file=True):
        """ This function formats json file in a compacter file format. It removes line break in aray
        """
        new_content=""
        # sleep for a second to give an upper stream function to store context into the file
        with (open(file_fpath, mode="r", encoding="utf-8") as file):
            run_strip_line = False

            for line in file:
                # skip run
                if  re.search(r"\[(.*?)\]", line):
                    new_content += line
                    continue

                # if found start strip
                if line.find(':[') > -1:
                    run_strip_line = True
                    new_content += line.rstrip()
                    continue

                # if found close strip
                if run_strip_line == True and line.find(']')>-1:
                    new_content += line.lstrip()
                    run_strip_line = False
                    continue

                if run_strip_line == True:
                    stripped_line = line.strip()
                    new_content += stripped_line
                else:
                    new_content += line

        if new_content != "":
            if modify_file == False:
                return new_content

            with (open(file_fpath, mode="w", encoding="utf-8") as file):
                file.write(new_content)

    def load_file(self, file_fpath):
        """
        """
        with open(file_fpath, "r", encoding="utf-8") as f:
            file_context = f.read()
            return file_context

    def save_file(self, path_to_file, file_context):
        """
        """
        with open(path_to_file, "w", encoding="utf-8") as f:
            f.write(file_context)
            return True

    def make_dirs(self, path_to_dir):
        if not os.path.isdir(path_to_dir):
            os.makedirs(path_to_dir)

    def deprecated_get_all_dir_files(self, dir_path, file_extension, file_names_only=False):
        """
        """

        # iterating over all files with determine extension
        result_arr = []

        for files in os.listdir(dir_path):
            if files.endswith(file_extension):
                if file_names_only == False:
                    result_arr.append(os.path.join(dir_path,files))
                else:
                    result_arr.append(files)
            else:
                continue

        return result_arr

    def get_file_names_with_extensions(self, dir_path, *file_extensions):
        """
        """

        # iterating over all files with determine extension
        result_arr = []

        for files in os.listdir(dir_path):
            for file_extension in file_extensions:
                if files.endswith(file_extension):
                        result_arr.append(files)
            else:
                continue

        return result_arr

    def get_all_dir_names(self, path_to_dir):
        """ """
        # iterating over all files with determine extension
        result_arr = []

        for dir in os.listdir(path_to_dir):
            result_arr.append(dir)

        return result_arr

    def get_specific_files(self, path_to_dir, file_name):
        """
        """

        # iterating over all files with determine extension
        result_arr = []
        for path in pathlib.Path(path_to_dir).rglob(file_name):
            result_arr.append(str(path))
        return result_arr

    def get_file_name(self, path_to_file):
        return os.path.basename(path_to_file)

    def list_dir(self, path_to_dir):
        """ It's a wrapper function for os.listdir
        """
        return os.listdir(path_to_dir)

    def is_file(self, path_to_file):
        return os.path.isfile(path_to_file)

    def is_dir(self, path_to_dir):
        return os.path.isdir(path_to_dir)

    def check_dir(self, path_to_dir):
        dir_exists = os.path.isdir(path_to_dir)
        files_in_dir = 0
        if dir_exists:
            files_in_dir = len(os.listdir(path_to_dir))

        return dir_exists, files_in_dir

    def cp_file(self, path_to_file, target_dir, target_file_name=None):

        file_exists = os.path.isfile(path_to_file)
        dir_exists = os.path.isdir(target_dir)

        if dir_exists == False:
            self.make_dirs(target_dir)

        if file_exists:
            destination = target_dir
            if target_file_name != None:
                destination = os.path.join(target_dir, target_file_name)
                shutil.copyfile(path_to_file, destination)
            else:
                shutil.copy(path_to_file, destination)

    def backup_file(self, path_to_file, target_dir=None):

        if target_dir is None:
            target_dir = os.path.dirname(path_to_file)
        base_fname = os.path.basename(path_to_file)
        target_file = base_fname + '.backup_' + datetime.now().strftime("%Y%m%d%H%M%S")
        self.cp_file(path_to_file, target_dir, target_file_name=target_file)

        logger.debug(f"The file {base_fname} has been backed up to the: {os.path.join(target_dir,target_file)}")

    def cleanup_dir(self, path_to_dir):
        """ This function remove all subfolders from path_to_dir. The folder path_to_dir will be not removed.

        :param path_to_dir:
        :return:
        """
        logger.debug(f"The files and folders from directory:{path_to_dir} will be removed")
        subfolders_are_not_exist = True
        files_are_not_exist = True
        for folder_name, subfolders, filenames in os.walk(path_to_dir):
            if len(subfolders)>0:
                subfolders_are_not_exist = False
                # clean up all subfolders in directory path
                logger.debug('Following subfolders are removed:')
                for subfolder in subfolders:
                    folder_to_remove = os.path.join(folder_name, subfolder)
                    shutil.rmtree(folder_to_remove)
                    logger.debug(folder_to_remove)
            else:
                for filename in filenames:
                    files_are_not_exist = False
                    file_to_remove = os.path.join(folder_name, filename)
                    os.remove(file_to_remove)

        if subfolders_are_not_exist:
            logger.debug('Directory has no subfolders')

        if files_are_not_exist:
            logger.debug('Directory has no files')

    def copy_file_without_comments(self, path_to_file, path_to_target_file, comment_sign='#'):
        """ This function copy templates file without comments to the specified pipeline
        """
        target_file = open(path_to_target_file, 'w')
        with open(path_to_file, "r") as file:
            for line in file:
                if line.find(comment_sign) < 0:
                    target_file.write(line)


        target_file.close()

    def load_combined_json(self, first_json_fpath, second_json_fpath) -> {}:
        """ combine two json files and return a dict
        """
        first_dict = {}
        second_dict = {}

        if first_json_fpath is not None and self.is_file(first_json_fpath):
            self.is_file(first_json_fpath)
            first_dict = self.load_json(first_json_fpath)

        if second_json_fpath is not None and self.is_file(second_json_fpath):
            second_dict = self.load_json(second_json_fpath)

        result_maping = self.dict_update(first_dict, second_dict)

        return result_maping

    def dict_update(self, dict_1, dict_2):
        """ Recursively update of a dict.
        """
        for key, value in dict_1.items():
            if key not in dict_2:
                dict_2[key] = value
            elif isinstance(value, dict):
                self.dict_update(value, dict_2[key])
        return dict_2

    def string_to_dict(self, input_string):
        result_json = ast.literal_eval('{'+input_string+'}')
        return dict(result_json)

    def unpack_archive(self, filename, extract_dir):
        shutil.unpack_archive(filename = filename, extract_dir=extract_dir)

    def gunzip_file(self, gz_fpath, cleanup_file=True):

        file_fpath = self.replace_file_extension(gz_fpath, '')
        try:
            with gzip.open(gz_fpath, 'rb') as f_in:
                with open(file_fpath, 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)
                    if cleanup_file == True:
                        os.remove(gz_fpath)  # delete zipped file

            return file_fpath

        except gzip.BadGzipFile as e:
            logger.error(e)
            return None

    def gzip_file(self, file_fpath, cleanup_file=True):

        gz_file_fpath = file_fpath + '.gz'

        try:
            with open(file_fpath, 'rb') as f_in:
                with gzip.open(gz_file_fpath, 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)
                    if cleanup_file == True:
                        os.remove(file_fpath)  # delete file

            return gz_file_fpath

        except gzip.BadGzipFile as e:
            logger.error(e)
            return None

    def gunzip_csv_files(self, gz_dpath, cleanup_file=True):
        gz_file_list = self.get_specific_files(gz_dpath, '*.csv.gz')
        for gz_fpath in gz_file_list:
            self.gunzip_file(gz_fpath, cleanup_file)

    def gzip_csv_files(self, gz_dpath, cleanup_file=True):
        gz_file_list = self.get_specific_files(gz_dpath, '*.csv')
        for gz_fpath in gz_file_list:
            self.gzip_file(gz_fpath, cleanup_file)

    def check_dict_key_exist(self, doc_dict: dict, key: str) -> bool:
        result = False
        key_list = list(doc_dict.keys())
        if(key_list.count(key) == 1):
            result = True

        return result

    def is_file_gzip(self, file_path):
        try:
            with open(file_path, 'rb') as test_f:
                if test_f.read(2) == b'\x1f\x8b':
                    return True
                else:
                    return False

        except  OSError as err:
            logger.error(f"The attempt to open file: {file_path} throws exception: {err}")

    def add_extension(self, file_path, extension):
        path_with_extension=file_path+extension
        os.rename(file_path, path_with_extension)
        return path_with_extension




# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

from petaly.utils.utils import measure_time
from petaly.utils.file_handler import FileHandler
import os
import sys
import csv

import mysql.connector


class MysqlConnector():

    def __init__(self, endpoint_attr):
        """ """
        self.connector_id = "mysql"
        self.metaquery_quote = ""

        self.conn = self.get_connection(endpoint_attr)
        self.database = endpoint_attr.get('database_name')


    def get_connection_dsn(self):
        return os.getenv("MYSQL_DSN")

    def compose_connection_params (self, endpoint_attr):
        connection_params = {"user": endpoint_attr.get('database_user'),
                             "password": endpoint_attr.get('database_password'),
                             "host": endpoint_attr.get('database_host'),
                             "port": endpoint_attr.get('database_port'),
                             "database": endpoint_attr.get('database_name'),
                             'allow_local_infile': True,
                             'use_pure': False
                             }
        return connection_params
    def get_connection(self, endpoint_attr):
        """
        """
        connection_params = self.compose_connection_params(endpoint_attr)
        try:
            conn = mysql.connector.connect(**connection_params)
            return conn

        except (mysql.connector.Error, IOError) as error:
            logger.error(error)
            sys.exit()

    def get_cursor(self):
        return self.conn.cursor(dictionary=True, buffered=False)

    def get_query_result(self, query):
        """ """
        try:

            with self.get_cursor() as cur:
                cur.execute(f"USE {self.database};")
                cur.execute(query)
                rows = cur.fetchall()
                logger.debug(f"Found {cur.rowcount} rows in query.")

                return rows

        except (mysql.connector.Error, IOError) as error:
            logger.error(error)
            sys.exit()

    def extract_to(self, extract_to_stmt, data_fpath, extract_options):

        self.extract_to_fetchone(extract_to_stmt, data_fpath, extract_options)

        # extract_to_fetchmany can be used too
        #self.extract_to_fetchmany(extract_to_stmt, data_fpath, batch_size=10000)
        ### deprecated
        #self.extract_to_fetchall(extract_to_stmt, data_fpath)


    def cleanup_linebreak_in_fields(self, row):
        """ This function cleans up a line terminator if it is present in a table field.
        This function is currently in testing mode and is used manually by adding a by adding a parameter to the pipeline.
        In the case of dict, the reference get passes, so no return is required.
        """
        for key, value in row.items():
            value = str(value).replace("\r", "").replace("\n", "")
            row.update({key:value})


    def extract_to_fetchone(self, extract_to_stmt, data_fpath, extract_options):
        """  """
        try:
            cur = self.get_cursor()
            cur.execute(extract_to_stmt)
            with (open(data_fpath, 'w') as file):

                row = cur.fetchone()

                csvwriter = csv.DictWriter(file,
                                           fieldnames=row,
                                           delimiter=extract_options.get("delimiter"),
                                           quotechar=extract_options.get("quotechar"),
                                           escapechar=extract_options.get("escapechar"),
                                           quoting=extract_options.get("quoting"),
                                           lineterminator=extract_options.get("lineterminator"),
                                          )

                if row is not None and extract_options.get("header"):
                    csvwriter.writeheader()

                while row is not None:

                    if extract_options.get("cleanup_linebreak_in_fields"):
                        self.cleanup_linebreak_in_fields(row)
                    csvwriter.writerow(row)
                    row = cur.fetchone()

                file.close()

            cur.close()

        except (mysql.connector.Error, IOError) as error:
            logger.debug(extract_to_stmt)
            logger.error(error)

    def describe_table(self, table_name):
        show_table_query = f"DESCRIBE {table_name}"
        with self.get_cursor() as cursor:
            cursor.execute(show_table_query)
            # Fetch rows from last executed query
            result = cursor.fetchall()
            for row in result:
                logger.debug(row)

    def extract_to_fetchmany(self, extract_to_stmt, data_fpath, extract_options, batch_size=10000):
        """ """
        try:
            cur = self.get_cursor()
            cur.execute(extract_to_stmt)

            with (open(data_fpath, 'w', newline='') as file):
                row = cur.fetchone()
                csvwriter = csv.DictWriter(file,
                                           fieldnames=row,
                                           delimiter=extract_options.get("delimiter"),
                                           quotechar=extract_options.get("quotechar"),
                                           escapechar=extract_options.get("escapechar"),
                                           quoting=extract_options.get("quoting"),
                                           lineterminator=extract_options.get("lineterminator"),
                                          )
                csvwriter.writeheader()
                csvwriter.writerow(row)

                rows = cur.fetchmany(size=batch_size)
                while len(rows) > 0:
                    csvwriter.writerows(rows)
                    rows = cur.fetchmany(size=batch_size)

                file.close()

            cur.close()

        except (mysql.connector.Error, IOError) as error:
            logger.debug(extract_to_stmt)
            logger.error(error)

    @measure_time
    def extract_to_fetchall(self, extract_to_stmt, data_fpath, extract_options):
        """  """
        try:
            cur = self.get_cursor()
            cur.execute(extract_to_stmt)

            with (open(data_fpath, 'w', newline='') as file):
                row = cur.fetchone()
                csvwriter = csv.DictWriter(file,
                                           fieldnames=row,
                                           delimiter=extract_options.get("delimiter"),
                                           quotechar=extract_options.get("quotechar"),
                                           escapechar=extract_options.get("escapechar"),
                                           quoting=extract_options.get("quoting"),
                                           lineterminator=extract_options.get("lineterminator"),
                                           )

                csvwriter.writeheader()
                csvwriter.writerow(row)
                rows = cur.fetchall()
                csvwriter.writerows(rows)

                file.close()

            cur.close()

        except (mysql.connector.Error, IOError) as error:
            logger.debug(extract_to_stmt)
            logger.error(error)

    def load_from(self, load_from_stmt):
        """  """
        try:
            cur = self.get_cursor()
            cur.execute(f"USE {self.database};")
            cur.execute(load_from_stmt)
            self.conn.commit()
        except (mysql.connector.Error, IOError) as error:
            logger.debug(load_from_stmt)
            logger.error(error)

    def drop_table(self, table_name):
        try:
            sql = f"DROP TABLE IF EXISTS {table_name}"
            cur = self.get_cursor()
            cur.execute(f"USE {self.database};")
            cur.execute(sql)  # Make an API request.
            self.conn.commit()
            logger.debug(f"Table {table_name} was dropped.")

        except (mysql.connector.Error, IOError) as error:
            logger.debug(sql)
            logger.error(error)

    def execute_sql(self, sql_stmt):
        try:
            cur = self.get_cursor()
            cur.execute(f"USE {self.database};")
            cur.execute(sql_stmt)  # Make an API request.
            self.conn.commit()
            logger.debug("Query was executed.")

        except (mysql.connector.Error, IOError) as error:
            logger.debug(sql_stmt)
            logger.error(error)

if __name__ == "__main__":
      pass



# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from petaly.utils.file_handler import FileHandler
from petaly.core.pipeline import Pipeline
from petaly.cli.cli_menu import CliMenu


class CliCleanup():

    def __init__(self, main_config):
        self.m_conf = main_config
        self.cli_menu = CliMenu(main_config)
        self.console = self.cli_menu.console
        self.f_handler = FileHandler()

    def cleanup_data_objects(self, pipeline_name, object_names):

        if pipeline_name is None:
            pipeline_name = self.cli_menu.force_assign_value(key='pipeline_name', message="Provide pipeline name. Pipeline with this name should already exists.")

        pipe = Pipeline(pipeline_name, self.m_conf)
        pipeline_all_obj = pipe.get_pipeline_entire_config()

        if object_names is None:
            object_names = self.cli_menu.force_assign_value(key='objects', message="Provide object name or comma separated list.")

        if type(object_names) == str:
            object_name_arr = [item.strip() for item in object_names.split(',')]

        if len(object_name_arr) > 0:
            self.remove_data_objects(pipeline_all_obj=pipeline_all_obj, object_name_arr=object_name_arr, pipeline_fpath=pipe.pipeline_fpath)
            self.console.print(f"Data-Objects were removed for pipeline {pipe.pipeline_name}. For further configuration review the yaml file: {pipe.pipeline_fpath} ")
        else:
            self.console.print(f"Data-Objects weren't specified for pipeline {pipe.pipeline_name}. For further configuration review the yaml file: {pipe.pipeline_fpath} ")

    def remove_data_objects(self, pipeline_all_obj, object_name_arr, pipeline_fpath):
        data_objects_spec = pipeline_all_obj[1].get('data_objects_spec')
        for object_name in object_name_arr:
            for idx, object_spec in enumerate(data_objects_spec):
                if object_name == object_spec.get('object_spec').get('object_name'):
                    data_objects_spec.pop(idx)

        pipeline_all_obj[1]['data_objects_spec'] = data_objects_spec

        self.f_handler.backup_file(pipeline_fpath)
        self.f_handler.save_dict_to_yaml(pipeline_fpath, pipeline_all_obj, dump_all=True)


# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

import os
import sys
import time
from petaly.core.f_extractor import FExtractor

class CsvExtractor(FExtractor):

    def __init__(self, pipeline):
        self.file_format = 'csv'
        super().__init__(pipeline)

    def extract_data(self):
        super().extract_data()

    def extract_to(self, extractor_obj_conf):
        """
        """
        # cleanup pipeline directory before run
        output_object_dpath = extractor_obj_conf.get('output_object_dpath')
        self.f_handler.cleanup_dir(output_object_dpath)

        object_name = extractor_obj_conf.get('object_name')
        object_source_dir = extractor_obj_conf.get('object_source_dir')
        file_names = extractor_obj_conf.get('file_names')

        if len(file_names) == 0 or file_names[0] is None:
            file_names = self.f_handler.get_file_names_with_extensions(object_source_dir, self.file_format)

        for file_name in file_names:
            file_source_fpath = os.path.join(object_source_dir, file_name)
            if self.f_handler.is_file(file_source_fpath):
                self.f_handler.cp_file(file_source_fpath, output_object_dpath)
            else:
                logger.error(f"The file: {file_source_fpath} wasn't found. Check the source and pipeline.yaml configuration.")
                return False

        first_file_fpath = os.path.join(output_object_dpath, file_names[0])

        logger.info(f"Check if the file {first_file_fpath} is compressed.")
        # check if file is gzip. if gzip unzip it
        if self.f_handler.is_file_gzip(first_file_fpath):
            if not self.f_handler.check_file_extension(first_file_fpath, '.gz'):
                first_file_fpath = self.f_handler.add_extension(first_file_fpath,'.gz')
            first_file_fpath = self.f_handler.gunzip_file(first_file_fpath, cleanup_file=True)

        # analyse file structure
        parquet_fpath = self.analyse_file_structure(first_file_fpath, object_name, file_format_extension= '.' + self.file_format)

        meta_table = self.compose_metadata_file(parquet_fpath, object_name)
        self.save_metadata_into_file(meta_table)

        # self.describe_parquet_metadata(parquet_fpath)



# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import logging.config
import os.path
#from datetime import datetime

from petaly.utils.file_handler import FileHandler


def setup_logging(logging_config_fpath, log_dir, logging_mode='INFO'):
    """
    """
    f_handler = FileHandler()
    config = f_handler.load_json(logging_config_fpath)
    if logging_mode == 'DEBUG':
        config['handlers']['console'].update({'level': logging_mode})
        config['handlers']['console'].update({'show_level': True})
        config['handlers']['console'].update({'omit_repeated_times': True})
        config['handlers']['console'].update({'show_path': True})
        config['handlers']['console'].update({'enable_link_path': True})
        config['handlers']['file'].update({'level': logging_mode})
        config['loggers']['petaly'].update({'level':logging_mode})
        config['root'].update({'level':logging_mode})
        config['formatters']['format_file'].update({"format":"%(asctime)s - %(name)s - %(levelname)s - %(message)s"})
        #config['formatters']['format_console'].update({"format": "%(name)s - %(message)s"})
        config['formatters']['format_console'].update({"format": "%(message)s"})

    log_fpath = os.path.join(log_dir, config.get('handlers').get('file').get('filename'))
    config['handlers']['file'].update({'filename': log_fpath})
    logging.config.dictConfig(config)




# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

import os
import sys
from rich.console import Console

from configparser import ConfigParser, ExtendedInterpolation
from petaly.utils.file_handler import FileHandler
from petaly.sysconfig.load_class import load_class_obj


class MainConfig:

    def __init__(self):
        self.f_handler = FileHandler()
        self.console = Console()
        self.pipeline_fname = 'pipeline.yaml'
        self.env_config_dpath = os.getenv('PETALY_CONFIG_DIR')
        self.main_config_fname = 'petaly.ini'
        self.logging_config_fname = 'logging_config.json'

        self.class_config_fname = 'class_config.json'
        self.pipeline_meta_config_fname = 'pipeline_meta_config.json'
        self.extractor_type_transformer_fname = 'extractor_type_transformer.json'
        self.type_mapping_fname = '{source_connector_id}.json'
        self.metadata_sql_fname = 'metadata.sql'
        self.extract_to_stmt_fname = 'extract_to_stmt.sql'
        self.load_from_stmt_fname = 'load_from_stmt.sql'
        self.create_table_stmt_fname = 'create_table_stmt.sql'
        self.connector_attributes_fname = 'connector_attributes.json'
        self.pipeline_outdated_arguments_fname = 'pipeline_outdated_arguments.json'

        self.base_dpath = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

        self.src_dpath = os.path.dirname(self.base_dpath)
        self.root_dpath = os.path.dirname(self.src_dpath)
        self.main_config_fpath = None

        self.sysconfig = os.path.join(self.base_dpath, 'sysconfig')
        self.sysconfig_config = os.path.join(self.sysconfig, 'config')

        self.templates_main_config_fpath = os.path.join(self.sysconfig_config, 'template_' + self.main_config_fname)
        self.logging_config_fpath = os.path.join(self.sysconfig_config, self.logging_config_fname)
        self.pipeline_meta_config_fpath = os.path.join(self.sysconfig_config, self.pipeline_meta_config_fname)
        self.class_sysconfig_fpath = os.path.join(self.sysconfig_config, self.class_config_fname)
        self.pipeline_outdated_arguments_fpath = os.path.join(self.sysconfig_config, self.pipeline_outdated_arguments_fname)

        self.workspace_config = {
                                            "pipeline_dir_path": None,
                                            "logs_dir_path": None,
                                            "output_dir_path": None
                                         }
        self.global_settings = {
                                        "logging_mode": "INFO"
                                        }


    def set_main_config_fpath(self, config_file_path, init_main_config=False):
        """
        """
        # 1. if config file wasn't pass, set to default one
        if config_file_path is None:
            config_file_path = self.main_config_fname

        # 2. exit, if config file is not absolute or env variable wasn't set
        if os.path.isabs(config_file_path) is False:
            if self.env_config_dpath:
                config_file_path = os.path.join(self.env_config_dpath, config_file_path)
            else:
                self.console.print(self.missing_main_config_file_message())
                sys.exit()

        # 3. exit, if file hasn't *.ini file extension
        if self.f_handler.check_file_extension(config_file_path, '.ini') is False:
            self.console.print(self.missing_main_config_file_message())
            sys.exit()

        # 4. If the configuration file or directory does not exist, create it.
        if self.f_handler.is_file(config_file_path) is False:
            if init_main_config == True:
                if self.f_handler.is_dir(config_file_path):
                    config_file_name = self.main_config_fname
                else:
                    config_file_name = os.path.basename(config_file_path)

                self.f_handler.cp_file(self.templates_main_config_fpath, os.path.dirname(config_file_path), os.path.basename(config_file_name))

                self.console.print(f"The main config file was created: {config_file_path}\n"
                      f"Open it with an editor and provide absolute paths for the following parameters: \nlogs_dir_path= \npipeline_dir_path= \noutput_dir_path=\n")
            else:
                self.console.print(self.missing_main_config_file_message())

        self.main_config_fpath = config_file_path

    def load_main_config_file(self):

        conf_parser = ConfigParser(interpolation=ExtendedInterpolation())
        conf_parser.read(self.main_config_fpath)

        return conf_parser

    def check_main_config_section(self, conf_parser, section_name):

        if conf_parser.has_section(section_name):
            return_result = True
        else:
            self.console.print(f"The section {section_name} in petaly.ini config file is not specified.")
            return_result = False

        return return_result

    def validate_workspace_abs_paths(self, conf_parser):

        return_value = True

        for key in self.workspace_config.keys():

            if key in conf_parser.options('workspace_config'):
                abs_path = conf_parser.get('workspace_config', key)
                if os.path.isabs(abs_path) is False:
                    self.console.print(f"The value of the option {key} is not specified or the path to the directory is not absolute.")
                    return_value = False
            else:
                self.console.print(f"The option {key} is not specified.")
                return_value = False

        return  return_value

    def set_workspace_dpaths(self):
        """
        """
        section_name = 'workspace_config'

        conf_parser = self.load_main_config_file()
        if self.check_main_config_section(conf_parser, section_name):
            if self.validate_workspace_abs_paths(conf_parser):
                self.pipeline_base_dpath = conf_parser.get(section_name,'pipeline_dir_path')
                self.logs_base_dpath = conf_parser.get(section_name,'logs_dir_path')
                self.output_base_dpath = conf_parser.get(section_name,'output_dir_path')
        else:
            self.console.print(f"Check petaly config file: {self.main_config_fpath}")
            sys.exit()

    def set_global_settings(self):

        section_name = 'global_settings'

        conf_parser = self.load_main_config_file()

        if self.check_main_config_section(conf_parser, section_name):

            for key in self.global_settings.keys():
                if key in conf_parser.options(section_name):
                    value = conf_parser.get(section_name, key)
                    if key == 'logging_mode':
                        if value in ('INFO', 'DEBUG'):
                            self.global_settings['logging_mode'] = value
                        else:
                            self.console.print(f"The option logging_mode supports INFO or DEBUG mode only. Check logging_mode under section global_settings in petaly.ini.")
                else:
                    self.console.print(f"The option {key} is not specified under section global_settings in petaly.ini.")

    def missing_main_config_file_message(self):
        return    (f"To initialize config file for the first time, provide the absolute path to petaly config file: init -c /ABSOLUTE_PATH_TO_PETALY_CONFIG_DIR/{self.main_config_fname}\n"
                   f"To skip '-c' argument at runtime, set an environment variable: export PETALY_CONFIG_DIR=/ABSOLUTE_PATH_TO_PETALY_CONFIG_DIR\n")


    def get_platform_attributes(self, platform_id):
        platforms_cl_config = self.f_handler.load_json(self.class_sysconfig_fpath).get('platforms')
        platform_config = platforms_cl_config.get(platform_id)

        if platform_config is None:
            logger.warning(f"The platform {platform_id} in class_config.json is not specified")
            sys.exit()

        return platform_config

    def get_supported_platforms(self, connector_id):
        platforms_cl_config = self.f_handler.load_json(self.class_sysconfig_fpath).get("connectors")
        platform_type_list = platforms_cl_config.get(connector_id).get('supported_platforms')
        if platform_type_list is None or len(platform_type_list)==0:
            platform_type_list = ['local']
        return platform_type_list

    def set_extractor_paths(self, connector_id):
        connector_dpath = self.get_connector_dpath(connector_id)
        self.connector_metadata_sql_fpath = os.path.join(connector_dpath, self.metadata_sql_fname)
        self.connector_extract_to_stmt_fpath = os.path.join(connector_dpath, 'config', self.extract_to_stmt_fname)
        return True

    def get_available_connectors(self):
        connector_type = (self.f_handler.load_json(self.class_sysconfig_fpath).get("connectors").keys())
        return connector_type

    def set_loader_paths(self, connector_id):
        connector_dpath = self.get_connector_dpath(connector_id)
        self.connector_load_from_stmt_fpath = os.path.join(connector_dpath, 'config', self.load_from_stmt_fname)
        self.connector_create_table_stmt_fpath = os.path.join(connector_dpath, 'config', self.create_table_stmt_fname)

        return True

    def get_connector_dpath(self, connector_id):
        """ The connector_id_dpath has a dot as path delimiter in class_config.json, e.g. connector_id_dpath: "connectors.mysql".
            To make it cross-platform compatible the split('.') and replace with directory delimiter is required.
        """
        connector_class_config = self.get_connector_class_config(connector_id)
        connector_id_dpath = connector_class_config.get('connector_dpath')
        connector_id_dpath_list = connector_id_dpath.split('.')
        connector_dpath = os.path.join(self.src_dpath, *connector_id_dpath_list)
        return connector_dpath

    def get_connector_class_config(self, connector_id):
        connectors_cl_config = self.f_handler.load_json(self.class_sysconfig_fpath).get("connectors")
        connector_class_config = connectors_cl_config.get(connector_id)

        if not connector_class_config:
            logger.warning(f"The connector_id {connector_id} in class_config.json is not specified")
            sys.exit()

        return connector_class_config

    def get_connector_attributes(self, connector_id):

        #connector_class_config = self.get_connector_class_config(connector_id)
        connector_dpath = self.get_connector_dpath(connector_id)
        connector_attributes_fpath = os.path.join(connector_dpath, 'config', self.connector_attributes_fname)

        return self.f_handler.load_json(connector_attributes_fpath)

    def get_type_mapping_path(self, connector_id, source_connector_id):

        type_mapping_path = self.get_connector_class_config(connector_id).get("type_mapping")
        type_mapping_fpath = os.path.join(self.src_dpath, *type_mapping_path.split('.'), self.type_mapping_fname)
        type_mapping_fpath = type_mapping_fpath.format(source_connector_id=source_connector_id)

        return type_mapping_fpath

    def get_extractor_type_transformer_fpath(self, connector_id):

        connector_dpath = self.get_connector_dpath(connector_id)
        return os.path.join(connector_dpath, 'config', self.extractor_type_transformer_fname)

    def get_extractor_class(self, connector_id):

        connector_class_config = self.get_connector_class_config(connector_id)
        return self.get_class_obj(connector_class_config, 'extractor')

    def get_loader_class(self, connector_id):
        connector_class_config = self.get_connector_class_config(connector_id)
        return self.get_class_obj(connector_class_config,'loader')

    def get_class_obj(self, connector_class_config, class_type):

        module_path = connector_class_config.get('connector_dpath') + '.' + connector_class_config.get(class_type).get('module_path')
        class_name = connector_class_config.get(class_type).get('class_name')
        class_object = load_class_obj(module_path, class_name)
        return class_object

    def get_connector_category(self, connector_id):
        connector_category = self.get_connector_class_config(connector_id).get('connector_category')
        return connector_category

    def get_pipeline_outdated_arguments(self):
        return self.f_handler.load_json(self.pipeline_outdated_arguments_fpath)


# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

import os
import sys
import logging

from petaly.utils.file_handler import FileHandler
from petaly.core.db_loader import DBLoader
from petaly.utils.utils import FormatDict
from petaly.connectors.gcp.bigquery.bq_connector import BQConnector
from petaly.connectors.gcp.gs.gs_connector import GSConnector


class BQLoader(DBLoader):
    def __init__(self, pipeline):
        self.db_connector = BQConnector()
        self.gs_connector = GSConnector()
        self.f_handler = FileHandler()
        super().__init__(pipeline)
        self.cloud_bucket_name = self.pipeline.target_attr.get('gcp_bucket_name')
        self.cloud_project_id = self.pipeline.target_attr.get('gcp_project_id')
        self.cloud_region = self.pipeline.target_attr.get('gcp_region')
        #self.cloud_service_account = self.pipeline.target_attr.get('gcp_service_account')

        self.cloud_bucket_path = self.gs_connector.bucket_prefix + self.cloud_bucket_name + '/'
        self.load_from_bucket = False if self.cloud_bucket_name is None else True


    #def execute_sql(self, create_table_stmt):
    #    self.db_connector.execute_sql(create_table_stmt)

    def load_data(self):
        super().load_data()

    def load_from(self, loader_obj_conf):

        object_name = loader_obj_conf.get('object_name')

        #bq_job_config = loader_obj_conf.get('load_from_stmt')

        #bq_job_config_dict = self.f_handler.string_to_dict(bq_job_config)

        table_id = self.get_table_id(loader_obj_conf.get('table_ddl_dict'))
        output_data_object_dir = loader_obj_conf.get('output_data_object_dir')

        if loader_obj_conf.get('recreate_destination_object') == True:
            self.drop_table(loader_obj_conf)
        self.create_table(loader_obj_conf)

        self.f_handler.gzip_csv_files(output_data_object_dir, cleanup_file=True)
        file_list = self.f_handler.get_specific_files(output_data_object_dir, '*.csv.gz')

        if self.load_from_bucket == True:
            self.gs_connector.delete_gs_folder(self.cloud_bucket_name, self.pipeline.pipeline_name + '/' + object_name)
            bucket_file_list = self.load_files_to_gs(file_list, self.cloud_bucket_name, self.pipeline.pipeline_name, object_name)

            if len(bucket_file_list) > 0:
                file_list = bucket_file_list
            else:
                logging.error(f"Files upload to bucket failed. Try upload from local path: ")

        bq_job_config_dict = loader_obj_conf.get('load_from_stmt')

        for path_to_data_file in file_list:
            self.db_connector.load_from(bq_job_config_dict, path_to_data_file, table_id, self.load_from_bucket, self.cloud_region)


    def compose_create_table_stmt(self, loader_obj_conf):

        table_ddl_dict = loader_obj_conf.get('table_ddl_dict')
        create_table_stmt = table_ddl_dict.get('create_table_stmt')
        create_table_stmt = create_table_stmt.format_map(FormatDict(schema_name=table_ddl_dict.get('schema_name'),
                                                                    table_name=table_ddl_dict.get('table_name'),
                                                                    column_datatype_list=table_ddl_dict.get('column_datatype_list'),
                                                                    partition_by='',
                                                                    cluster_by='',
                                                                    table_options=''))
        loader_obj_conf.get('table_ddl_dict').update({'create_table_stmt': create_table_stmt})

        return loader_obj_conf
    def get_table_id(self, table_ddl_dict):

        return (f"{self.cloud_project_id}."
                f"{table_ddl_dict.get('schema_name')}."
                f"{table_ddl_dict.get('table_name')}")

    def drop_table(self, loader_obj_conf: dict):

        table_id = self.get_table_id(loader_obj_conf.get('table_ddl_dict'))
        self.db_connector.drop_table(table_id)

    def create_table(self, loader_obj_conf: dict):
        loader_obj_conf = self.compose_create_table_stmt(loader_obj_conf)
        self.f_handler.save_file(loader_obj_conf.get('table_ddl_dict').get('create_table_stmt_fpath'),
                                 loader_obj_conf.get('table_ddl_dict').get('create_table_stmt'))

        self.db_connector.execute_sql(loader_obj_conf.get('table_ddl_dict').get('create_table_stmt'))

    def compose_from_options(self):
        """
        """
        load_options = {}

        object_default_settings = self.pipeline.data_attributes.get("object_default_settings")
        columns_delimiter = object_default_settings.get("columns_delimiter")

        load_options.update({'delimiter': columns_delimiter})
        if columns_delimiter == "\t":
            load_options.update({'delimiter': '\\t'})

        header = True if object_default_settings.get("header") is None or True else False
        load_options.update({'header': header})

        # 3. OPTIONALLY ENCLOSED BY
        columns_quote = object_default_settings.get("columns_quote")
        quote_char = None
        if columns_quote == 'double':
            quote_char = '"'
        elif columns_quote == 'single':
            quote_char = "'"
        load_options.update({'quote_char': quote_char})

        return load_options

    def compose_load_from_stmt(self, data_object, loader_obj_conf):
        """ Its compose a copy from statement """
        load_data_options = self.compose_from_options()
        bq_load_from_stmt_fpath = self.f_handler.replace_file_extension(self.connector_load_from_stmt_fpath,'.json')
        load_from_stmt = self.f_handler.load_json(bq_load_from_stmt_fpath)

        #column_list = loader_obj_conf.get('table_ddl_dict').get('column_list')
        max_bad_records = 0
        skip_leading_rows = 1 if load_data_options.get("header") is True else 0
        field_delimiter = load_data_options.get("delimiter")
        quote_char=load_data_options.get("quote_char")
        load_from_stmt.update({"source_format": self.db_connector.bq_source_format})
        load_from_stmt.update({"skip_leading_rows": skip_leading_rows})
        load_from_stmt.update({'autodetect': False})
        load_from_stmt.update({'max_bad_records': max_bad_records})
        load_from_stmt.update({'field_delimiter': field_delimiter})
        load_from_stmt.update({'quote_character': quote_char})

        load_from_file_fpath = loader_obj_conf.get('load_from_stmt_fpath')
        load_from_file_fpath = self.f_handler.replace_file_extension(load_from_file_fpath, '.json')
        loader_obj_conf.update({'load_from_stmt_fpath':load_from_file_fpath})
        self.f_handler.save_dict_to_json(load_from_file_fpath, load_from_stmt)

        return load_from_stmt

    def load_files_to_gs(self, local_file_list, cloud_bucket_name, pipeline_name, object_name):
        """ upload file to GS bucket
        """
        bucket_file_list = []
        for file_local_fpath in local_file_list:
            file_name = os.path.basename(file_local_fpath)

            blob_path = pipeline_name + '/' + object_name + '/' + file_name
            self.gs_connector.upload_blob(file_local_fpath, cloud_bucket_name, blob_path)

            full_blob_path = self.gs_connector.bucket_prefix + cloud_bucket_name + '/' + blob_path

            bucket_file_list.append(full_blob_path)

            logging.debug(f"File upload: {full_blob_path}")

        return bucket_file_list


# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import rich.prompt as prompt
from rich.console import Console

from petaly.utils.file_handler import FileHandler
from collections import Counter, OrderedDict


class OrderedCounter(Counter, OrderedDict):
    'Counter that remembers the order elements are first seen'
    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__,
                           OrderedDict(self))

    def __reduce__(self):
        return self.__class__, (OrderedDict(self),)

class CliMenu():

    def __init__(self, main_config):
        self.use_pipeline_wizard = True
        self.console = Console()
        self.prompt = prompt
        self.m_conf = main_config
        self.f_handler = FileHandler()
        self.break_line = '------------------------------------------------------------------'

        self.pipeline_meta_config = self.f_handler.load_json(self.m_conf.pipeline_meta_config_fpath)

        self.composed_pipeline_config = [
                                            {'pipeline':
                                                 {'pipeline_attributes': {},
                                                  'source_attributes': {},
                                                  'target_attributes': {},
                                                  'data_attributes': {
                                                      'data_objects_spec_mode':{}
                                                  }
                                                  }
                                            },
                                            {
                                              'data_objects_spec': []
                                            }
                                        ]

    def force_assign_value(self, key, message):
        while True:
            self.console.print(message)
            value = prompt.Prompt.ask(f"[bold green]{key}[/bold green]")
            message = f"[red]{message}[/red]"
            if value.strip() != '':

                break
        return value

    def compose_pipeline(self, pipeline_name):


        self.compose_pipeline_attributes(pipeline_name)
        self.compose_endpoint_attributes('source_attributes')
        self.compose_endpoint_attributes('target_attributes')
        self.compose_data_attributes()

    def compose_pipeline_attributes(self, pipeline_name):
        """
        """
        self.use_pipeline_wizard = prompt.Confirm.ask("\nUse Pipeline wizard")

        predefined_values = {'pipeline_name': pipeline_name}
        pipeline_attributes = self.pipeline_meta_config.get("pipeline_attributes")
        assigned_attributes = self.assign_attributes(pipeline_attributes, predefined_values=predefined_values)
        self.composed_pipeline_config[0]['pipeline']['pipeline_attributes'].update(assigned_attributes)

    def compose_endpoint_attributes(self, endpoint_attributes_name):
        self.console.print(f"\n[bold]{self.break_line}[/bold]")
        self.console.print(f"Specify [bold]{endpoint_attributes_name}[/bold]\n")

        endpoint_attributes_dict = {}
        predefined_values = {}

        # step 1. specify endpoint type
        available_connectors = self.m_conf.get_available_connectors()
        connector_type = prompt.Prompt.ask(f"Specify [bold yellow]{endpoint_attributes_name}[/bold yellow] connector type", choices=available_connectors)
        predefined_values.update({'connector_type': connector_type})

        assigned_endpoint_attributes = self.assign_attributes(endpoint_attributes_dict, predefined_values=predefined_values)
        self.composed_pipeline_config[0]['pipeline'][endpoint_attributes_name].update(assigned_endpoint_attributes)

        # step 2. get connector category
        connector_category = self.m_conf.get_connector_class_config(connector_type).get('connector_category')

        # step 3. based on connector category define database or file parameters
        connector_attributes = self.m_conf.get_connector_attributes(connector_type)

        exclude_key_list = [None]
        if connector_category in ('file','storage') and endpoint_attributes_name == 'source_attributes':
            exclude_key_list = ['destination_file_dir', 'destination_blob_dir']

        assigned_connector_attributes = self.assign_attributes(connector_attributes, exclude_key_list=exclude_key_list, predefined_values=None)
        self.composed_pipeline_config[0]['pipeline'][endpoint_attributes_name].update(assigned_connector_attributes)

        # step 4. get and define platform type
        platform_type_list = self.m_conf.get_supported_platforms(connector_type)

        if len(platform_type_list) == 1:
            platform_type = platform_type_list[0]
        else:
            platform_type = prompt.Prompt.ask(f"Specify [bold green]platform_type[/bold green]",
                                              choices=platform_type_list)

        # step 5. assign platform type and platform attributes
        if platform_type != 'local':
            predefined_values.update({'platform_type': platform_type})
            platform_attributes = self.m_conf.get_platform_attributes(platform_id=platform_type)
            assigned_platform_attributes = self.assign_attributes(platform_attributes, predefined_values=predefined_values, exclude_key_list=['connector_type'])
            self.composed_pipeline_config[0]['pipeline'][endpoint_attributes_name].update(assigned_platform_attributes)

    def compose_data_attributes(self):

        data_attributes = self.pipeline_meta_config.get('data_attributes')

        self.console.print(f"\n[bold]{self.break_line}[/bold]")
        self.console.print(f"[bold]Specify default object settings[/bold]")

        object_default_settings = data_attributes.get('object_default_settings')
        assigned_object_default_settings = self.assign_attributes(object_default_settings, predefined_values=None)
        self.composed_pipeline_config[0]['pipeline']['data_attributes'].update({"object_default_settings": assigned_object_default_settings})

        self.console.print(f"\n[bold]{self.break_line}[/bold]")
        self.console.print(f"[bold]Specify data object attributes[/bold]")

        assigned_data_attributes = self.assign_attributes(data_attributes, predefined_values=None, exclude_key_list=[None])
        self.composed_pipeline_config[0]['pipeline']['data_attributes'].update(assigned_data_attributes)

    def compose_object_spec(self, pipeline, object_name, use_pipeline_wizard):

        self.use_pipeline_wizard = use_pipeline_wizard
        data_objects_spec = self.pipeline_meta_config.get('data_objects_spec')
        exclude_key_list = []
        connector_category = self.m_conf.get_connector_category(pipeline.source_attr.get('connector_type'))

        # exclude params for file load (csv, etc..)
        if connector_category not in ('file','storage'):
            exclude_key_list.append('object_source_dir')
            exclude_key_list.append('file_names')

        if object_name is None:
            object_name = self.force_assign_value(key='object_name',
                                                  message=data_objects_spec.get('object_name').get('key_comment'))

        exclude_key_list.append('object_name')
        predefined_values = {'object_name': object_name}

        assigned_attributes = self.assign_attributes(data_objects_spec, predefined_values=predefined_values, exclude_key_list=exclude_key_list)
        tmp_assigned_attributes = {}
        tmp_assigned_attributes.update({'object_spec': assigned_attributes})

        return tmp_assigned_attributes

    def assign_attributes(self, spec_attributes, predefined_values=None, exclude_key_list=None) -> dict:
        assigned_attributes = {}

        if predefined_values is not None:
            assigned_attributes.update(predefined_values)

        # cleanup dict from exclude keys
        if exclude_key_list is not None:
            for key in exclude_key_list:
                if self.f_handler.check_dict_key_exist(spec_attributes, key):
                    spec_attributes.pop(key)

        for key, value in spec_attributes.items():

            in_use = False if value.get('in_use') is None else value.get('in_use')

            if in_use:
                # 1. handle predefined_values
                if predefined_values is not None:
                    if self.f_handler.check_dict_key_exist(predefined_values, key):
                        assigned_value = predefined_values.get(key)
                        assigned_attributes.update({key: assigned_value})
                        continue

                # 2. define default and preassigned_values
                preassigned_values = value.get('preassigned_values')
                default_value =  None if value.get('default_value') is None else value.get('default_value')
                preassigned_values = None if preassigned_values[0] is None else preassigned_values
                assigned_value = default_value

                # 3. compose key comment and default value
                console_message = "\n"

                console_message += f"{value.get('key_comment')}"
                if default_value is not None:
                    console_message += f"Default: [bold blue]{default_value}[/bold blue]"

                self.console.print(console_message)

                if key == 'database_password':
                    if self.use_pipeline_wizard:
                        assigned_value = prompt.Prompt.ask(f"[bold green]{key}[/bold green]", password=True)

                elif value.get('key_type') == 'Integer':
                    if self.use_pipeline_wizard:
                        assigned_value = prompt.IntPrompt.ask(f"[bold green]{key}[/bold green]", default=default_value,
                                                              show_default=False)

                elif value.get('key_type') == 'Array':

                    if self.use_pipeline_wizard:
                        assigned_value = prompt.Prompt.ask(f"[bold green]{key}[/bold green]", default=default_value,
                                                           show_default=False)

                        if type(assigned_value) == str:
                            assigned_value = [item.strip() for item in assigned_value.split(',')]

                    if assigned_value is None:
                        assigned_value = [None]
                else:
                    if self.use_pipeline_wizard:
                        assigned_value = prompt.Prompt.ask(f"[bold green]{key}[/bold green]", choices=preassigned_values,
                                                           default=default_value, show_default=False)

                    if value.get('key_type') == 'Boolean':
                        assigned_value = True if assigned_value == 'true' else False

                assigned_attributes.update({key: assigned_value})

        return assigned_attributes


# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

import os
import logging

from petaly.connectors.gcp.bigquery.bq_connector import BQConnector
from petaly.connectors.gcp.gs.gs_connector import GSConnector
from petaly.core.db_extractor import DBExtractor
from petaly.utils.utils import FormatDict


class BQExtractor(DBExtractor):
    def __init__(self, pipeline):
        self.db_connector = BQConnector()
        self.gs_connector = GSConnector()

        super().__init__(pipeline)
        self.cloud_bucket_name = self.pipeline.source_attr.get('gcp_bucket_name')
        self.cloud_project_id = self.pipeline.source_attr.get('gcp_project_id')
        self.cloud_region = self.pipeline.source_attr.get('gcp_region')
        #self.cloud_service_account = self.pipeline.source_attr.get('gcp_service_account')

    def extract_data(self):
        super().extract_data()

    def get_query_result(self, meta_query):
        query_result = self.db_connector.get_metadata_result(meta_query)
        return query_result

    def extract_to(self, extractor_obj_conf):

        object_name = extractor_obj_conf.get('object_name')
        self.gs_connector.delete_gs_folder(self.cloud_bucket_name, object_name)

        # run export data
        extract_to_stmt = extractor_obj_conf.get('extract_to_stmt')
        extract_to_dict = self.f_handler.string_to_dict(extract_to_stmt)
        table_ref = extract_to_dict.get('table_ref')
        destination_uri = extract_to_dict.get('destination_uri')

        self.db_connector.extract_to(table_ref, destination_uri, self.cloud_region)

        output_file_dir = os.path.dirname(extractor_obj_conf.get('output_object_fpath'))

        blob_prefix = (self.pipeline.pipeline_name +'/'+ object_name).strip('/')
        # download export from bucket into local folder
        downloaded_file_list = self.gs_connector.download_files_from_bucket(
                                                    self.cloud_bucket_name,
                                                    blob_prefix,
                                                    specific_file_list=None,
                                                    destination_directory=output_file_dir)

        logging.debug(f"Following file list were downloaded from bucket:\n{downloaded_file_list}")

    def compose_extract_to_stmt(self, extract_to_stmt, extractor_obj_conf) -> dict:
        """ Its save copy statement into file
        """
        project_id = self.cloud_project_id
        object_name = extractor_obj_conf.get('object_name')
        dataset_id = extractor_obj_conf.get('source_schema_name')
        table_name = extractor_obj_conf.get('source_object_name')

        destination_blob_name = f"{self.pipeline.pipeline_name}/{object_name}/{object_name}_*.csv"
        destination_uri = f"{self.gs_connector.bucket_prefix }{self.cloud_bucket_name}/{destination_blob_name}"
        table_ref = f"{project_id}.{dataset_id}.{table_name}"

        extract_to_stmt = extract_to_stmt.format_map(FormatDict(table_ref=table_ref, destination_uri=destination_uri))

        return extract_to_stmt



# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)
import csv
from petaly.connectors.mysql.mysql_connector import MysqlConnector
from petaly.core.db_extractor import DBExtractor
from petaly.utils.utils import FormatDict



class MysqlExtractor(DBExtractor):
    def __init__(self, pipeline):
        self.db_connector = MysqlConnector(pipeline.source_attr)
        super().__init__(pipeline)

    def extract_data(self):
        super().extract_data()

    def get_query_result(self, meta_query):
        return self.db_connector.get_query_result(meta_query)

    def compose_extract_options(self, extractor_obj_conf) -> dict:
        """
        """
        object_settings = extractor_obj_conf.get('object_settings')

        extract_options = {}

        extract_options.update({"delimiter": object_settings.get("columns_delimiter")})

        columns_quote = object_settings.get("columns_quote")
        extract_options.update({"quoting": csv.QUOTE_ALL})
        if columns_quote == 'double':
            extract_options.update({"quotechar": "\""})
        elif columns_quote == 'single':
            extract_options.update({"quotechar": "\'"})
        else:
            extract_options.update({"quotechar": None})
            extract_options.update({"quoting": csv.QUOTE_NONE})

        extract_options.update({"escapechar": "\\"})
        extract_options.update({"lineterminator": "\n"})

        extract_options.update({"header": object_settings.get("header")})

        cleanup_linebreak_in_fields = object_settings.get("cleanup_linebreak_in_fields")
        extract_options.update({"cleanup_linebreak_in_fields": cleanup_linebreak_in_fields})

        return extract_options

    def extract_to(self, extractor_obj_conf):
        """
        """
        output_object_fpath = extractor_obj_conf.get("output_object_fpath")
        extract_to_stmt = extractor_obj_conf.get("extract_to_stmt")
        extract_options = self.compose_extract_options(extractor_obj_conf)
        logger.debug(f"Output File: {output_object_fpath}")
        logger.debug(f"Statement to execute:\n{extract_to_stmt}")

        self.db_connector.extract_to(extract_to_stmt, output_object_fpath, extract_options)

    def compose_extract_to_stmt(self, extract_to_stmt, extractor_obj_conf) -> dict:
        """ Its save copy statement into file
        """
        column_list = extractor_obj_conf.get("column_list")
        schema_name = extractor_obj_conf.get("source_schema_name")
        table_name = extractor_obj_conf.get("source_object_name")

        extract_to_stmt = extract_to_stmt.format_map(
        					FormatDict(column_list=column_list, schema_name=schema_name, table_name=table_name, null_as=''))

        return extract_to_stmt



# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

import time
from abc import ABC, abstractmethod

from petaly.core.composer import Composer
from petaly.utils.file_handler import FileHandler

from petaly.core.data_object import DataObject

class FLoader(ABC):

    def __init__(self, pipeline):
        self.pipeline = pipeline
        self.composer = Composer()
        self.f_handler = FileHandler()
        pass

    @abstractmethod
    def load_from(self, loader_obj_conf):
        pass

    def load_data(self):

        logger.info(f"[--- Load into {self.pipeline.target_connector_id} ---]")
        start_total_time = time.time()

        if self.pipeline.data_attributes.get("data_objects_spec_mode") == 'only':
            object_list = self.pipeline.data_objects
        else:
            #object_list = self.f_handler.get_all_dir_names(self.pipeline.output_pipeline_dpath)
            object_list = self.composer.get_object_list_from_output_dir(self.pipeline)

        for object_name in object_list:

            logger.info(f"Load object: {object_name} started...")
            start_time = time.time()

            loader_obj_conf = {}
            loader_obj_conf.update({'object_name': object_name})

            output_metadata_object_dir = self.pipeline.output_object_metadata_dpath.format(object_name=object_name)
            loader_obj_conf.update({'output_metadata_object_dir': output_metadata_object_dir})

            output_data_object_dir = self.pipeline.output_object_data_dpath.format(object_name=object_name)
            loader_obj_conf.update({'output_data_object_dir': output_data_object_dir})

            output_load_from_stmt_fpath = self.pipeline.output_load_from_stmt_fpath.format(object_name=object_name)
            loader_obj_conf.update({'load_from_stmt_fpath': output_load_from_stmt_fpath})

            self.load_from(loader_obj_conf)

            end_time = time.time()
            logger.info(f"Load object: {object_name} completed | time: {round(end_time - start_time, 2)}s")

        end_total_time = time.time()
        logger.info(f"Load completed, duration: {round(end_total_time - start_total_time, 2)}s")

    def get_data_object(self, object_name):
        return DataObject(self.pipeline, object_name)



# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

import os
import logging

from petaly.connectors.gcp.gs.gs_connector import GSConnector
from petaly.core.f_extractor import FExtractor
from petaly.utils.file_handler import FileHandler
from petaly.utils.utils import FormatDict

class GSExtractor(FExtractor):
    def __init__(self, pipeline):
        self.gs_connector = GSConnector()
        super().__init__(pipeline)
        self.cloud_bucket_name = self.pipeline.source_attr.get('gcp_bucket_name')
        self.cloud_project_id = self.pipeline.source_attr.get('gcp_project_id')
        self.cloud_region = self.pipeline.source_attr.get('gcp_region')

        self.f_handler = FileHandler()

    def extract_data(self):
        super().extract_data()

    def extract_to(self, extractor_obj_conf):

        output_object_dpath = extractor_obj_conf.get('output_object_dpath')
        self.f_handler.cleanup_dir(output_object_dpath)
        #object_name = extractor_obj_conf.get('object_name')
        #data_object = super().get_data_object(object_name)

        object_source_dir = extractor_obj_conf.get('object_source_dir')
        blob_prefix = object_source_dir.strip('/')
        file_names = extractor_obj_conf.get('file_names')
        output_file_dir = os.path.dirname(extractor_obj_conf.get('output_object_fpath'))

        file_names_full_path = []
        for file_name in file_names:
            file_names_full_path.append(f"{blob_prefix}/{file_name}")

        # download export from bucket into local folder
        downloaded_file_list = self.gs_connector.download_files_from_bucket(
                                                    self.cloud_bucket_name,
                                                    blob_prefix,
                                                    specific_file_list=file_names_full_path,
                                                    destination_directory=output_file_dir)

        logging.debug(f"Following file list were downloaded from bucket:\n{downloaded_file_list}")





# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

import os
from petaly.utils.file_handler import FileHandler

class TypeMapping():
    def __init__(self, pipeline):
        self.pipeline = pipeline
        self.f_handler = FileHandler()
        self.m_conf = pipeline.m_conf

    def get_type_mapping(self):
        """ get target-source type mapping. Return the first founded.

        """
        type_mapping_fpath = self.pipeline.pipeline_type_mapping_fpath.format(source_connector_id=self.pipeline.source_connector_id)
        if not self.f_handler.is_file(type_mapping_fpath):
            type_mapping_fpath = self.m_conf.get_type_mapping_path(self.pipeline.target_connector_id, self.pipeline.source_connector_id)

        logger.debug(f"Load data type mapping from: {type_mapping_fpath}")
        type_mapping_dict = self.f_handler.load_json(type_mapping_fpath)
        return type_mapping_dict


    def get_extractor_type_transformer(self):

        extractor_type_transformer_fpath = self.m_conf.get_extractor_type_transformer_fpath(self.pipeline.source_connector_id)
        type_mapping_dict = self.f_handler.load_json(extractor_type_transformer_fpath)
        if self.f_handler.is_file(self.pipeline.pipeline_extract_type_transformer_fpath):
            type_mapping_dict = self.f_handler.load_json(self.pipeline.pipeline_extract_type_transformer_fpath)
        return type_mapping_dict





# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys

logger = logging.getLogger(__name__)

import os
import time
import logging
from google.cloud import storage, exceptions
from petaly.utils.file_handler import FileHandler


class GSConnector():

    def __init__(self):
        self.connector_id = 'gs'
        self.bucket_prefix = 'gs://'
        self.f_handler = FileHandler()
        pass


    def upload_blob(self, full_fpath, bucket_name, destination_blob_name):
        """Function uploads a file to the GS bucket.
        """
        logging.debug(
            f"Load data from the local path {full_fpath} to the backet: {bucket_name} with destination path {destination_blob_name}; ")

        try:
            st_client = storage.Client()
            bucket = st_client.bucket(bucket_name)
            blob = bucket.blob(destination_blob_name)
            blob.upload_from_filename(full_fpath)
        except exceptions.GoogleCloudError as err:
            logging.error(err)
            pass

    def rename_blob(self, bucket_name, blob_file_name, new_blob_name = None):
        """ Function renamed blob by giving time sequence suffix in millisecond.

        ToDo: It should be reviewed as keeping blob in the same folder with the same name will cause load to bigquery
        """
        file_sufix = str(time.time_ns())

        if new_blob_name is None:
            new_blob_name = blob_file_name + "." + file_sufix

        logging.debug(f"Bucket renaming started: current name {blob_file_name} new name {new_blob_name}")

        try:
            st_client = storage.Client()
            bucket = st_client.bucket(bucket_name)
            blob = bucket.blob(blob_file_name)
            new_target_file = bucket.rename_blob(blob, new_blob_name)
            logging.debug(f"Blob {blob.name} has been renamed to {new_target_file.name}")

        except exceptions.GoogleCloudError as err:
            logging.error(err)

    def delete_gs_folder(self, bucket_name, folder_name):
        """
        """
        logging.debug(f"Delete folder in Google Storage: {bucket_name}/{folder_name}")
        try:
            st_client = storage.Client()
            bucket = st_client.get_bucket(bucket_name)
            """Delete object under folder"""
            blobs = list(bucket.list_blobs(prefix=folder_name))
            bucket.delete_blobs(blobs)
            logging.debug(f"Folder deleted: {bucket_name}/{folder_name}")
        except exceptions.GoogleCloudError as err:
            logging.error(err)
            pass

    def delete_gs_blob(self, bucket_name, blob_name):
        """ Function drop a blob in specific bucket
        """
        logging.debug(f"Blob {blob_name} in {bucket_name} will be deleted")
        try:
            st_client = storage.Client()
            bucket = st_client.bucket(bucket_name)
            bucket.delete_blob(blob_name)
            logging.debug(f"Blob {blob_name} in {bucket_name} has been deleted")
        except exceptions.GoogleCloudError as err:
            logging.error(err)
            pass

    def download_files_from_bucket(self, bucket_name, blob_prefix, specific_file_list, destination_directory):
        """
        """
        logging.debug(f"Download files from bucket-name: {bucket_name}; blob_prefix: {blob_prefix}; destination_directory: {destination_directory}")
        delimiter = '/'
        blob_prefix += delimiter
        download_specific_files_only = True if specific_file_list is not None and specific_file_list[0] is not None else False

        try:

            storage_client = storage.Client()
            bucket = storage_client.get_bucket(bucket_name)
            blobs = storage_client.list_blobs(bucket_name, prefix=blob_prefix, delimiter=delimiter)
            blob_result_list = []
            for blob in blobs:
                if download_specific_files_only:
                    if blob.name in (specific_file_list):
                        blob_result_list.append(blob.name)
                else:
                    blob_result_list.append(blob.name)

            downloaded_file_list = []
            for blob_fpath in blob_result_list:
                blob = bucket.blob(blob_fpath)
                extract_file_name = blob_fpath.rsplit('/')[-1]

                target_fpath = os.path.join(destination_directory, extract_file_name)

                if not self.f_handler.check_file_extension(target_fpath, '.gz'):

                    target_fpath += '.gz'

                blob.download_to_filename(filename=target_fpath)
                downloaded_file_list.append(target_fpath)

            return downloaded_file_list

        except exceptions.GoogleCloudError as err:
            logging.error(err)


# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

import os
import time
import logging
from google.cloud import exceptions, bigquery


class BQConnector():

    def __init__(self):
        self.connector_id = 'bigquery'
        self.metaquery_quote = ''
        self.bq_source_format = 'bigquery.SourceFormat.CSV'
        pass

    def extract_to(self, table_ref, destination_uri, region):

        logging.debug(f"Extract table {table_ref} to {destination_uri} started")
        try:
            bq_client = bigquery.Client()
            job_config = bigquery.job.ExtractJobConfig()
            job_config.compression = bigquery.Compression.GZIP

            extract_job = bq_client.extract_table(
                table_ref,
                destination_uri,
                location=region,
                job_config=job_config,
            )

            result = extract_job.result()  # Waits for job to complete.

            logging.debug(f"Table {table_ref} was loaded to {destination_uri}. Result: {result}")

        except exceptions.GoogleCloudError as err:
            logging.error(err)

    def execute_sql(self, query):
        """
        """
        logging.debug(f"Execute query.")
        try:
            bq_client = bigquery.Client()
            job = bq_client.query(query)  # API request.
            job.result()  # Waits for the query to finish.
            logging.debug(f"New table {job.destination.project}.{job.destination.dataset_id}.{job.destination.table_id} is created.")
        except exceptions.GoogleCloudError as err:
            logging.debug(query)
            logging.error(err)
            pass


    def get_metadata_result(self, metadata_query):
        logging.debug(f"Execute meta query.")
        try:
            bq_client = bigquery.Client()
            query_result = bq_client.query_and_wait(metadata_query)
            result_arr = []
            for row in query_result:
                result_arr.append(dict(list(row.items())))

            return result_arr
        except exceptions.GoogleCloudError as err:
            logging.debug(metadata_query)
            logging.error(err)
            pass

    def create_table_from_json_schema(self, table_schema_fpath, table_id):
        """ ToDo this function potentially for bq_connector
        """
        # TODO(dev): Change the table_id variable to the full name of the table you want to get schema from.
        logging.debug(f"Create table {table_id}.")
        try:
            bq_client = bigquery.Client()
            schema = bq_client.schema_from_json(table_schema_fpath)
            table = bigquery.Table(table_id, schema=schema)
            table = bq_client.create_table(table)  # API request
            logging.debug( f"Table {table_id} was created.")
        except Exception as e:
            logging.error(e)
            pass

        return table

    def load_from(self, bq_job_config, data_fpath, table_id, load_from_bucket, region):
        """ Load file from local file system or from bucket to BigQuery. The param load_from_bucket determine the behaviour.
        """

        logging.debug(f"Load data from the file {data_fpath}")

        try:
            bq_client = bigquery.Client()


            job_config = bigquery.LoadJobConfig(**bq_job_config)

            #job_config = bigquery.LoadJobConfig(source_format=bigquery.SourceFormat.CSV,
            #                                    skip_leading_rows=1,
            #                                    autodetect=False,
            #                                    max_bad_records=0,
            #                                    field_delimiter=";",
            #                                    quote_character='"')
            #

            rows_start = bq_client.get_table(table_id).num_rows  # Make an API request.

            if load_from_bucket:
                job = bq_client.load_table_from_uri(data_fpath,
                                                    table_id,
                                                    job_config=job_config,
                                                    location=region)
            else:
                with open(data_fpath, "rb") as source_file:
                    job = bq_client.load_table_from_file(source_file,
                                                         table_id,
                                                         job_config=job_config,
                                                         location=region)

            job.result()  # Waits for the job to complete.
            table = bq_client.get_table(table_id)  # Make an API request.
            rows_end = table.num_rows
            rows_loaded = rows_end - rows_start
            logging.debug(f"Loaded {rows_loaded} rows and {len(table.schema)} columns to {table_id} from file {data_fpath}")

        except exceptions.GoogleCloudError as err:
            logging.error(err)


    def drop_table(self, table_id):
        """ Function drop BigQuery table
        """
        logging.debug(f"DROP table if exist {table_id}")
        try:
            bq_client = bigquery.Client()
            bq_client.delete_table(table_id, not_found_ok=True)  # Make an API request.
            logging.debug(f"Table {table_id} was dropped.")
        except exceptions.GoogleCloudError as err:
            logging.error(err)
            pass
